#ifndef GDKFB_H
#define GDKFB_H 1

/**
 * PAVEL_MAGIC_X:
 *
 * Horizontal number of pixels represented by one text character.
 *
 * There is no real way how to handle all the character-size alignment during
 * GDK drawing if using non-1 value as GDK uses various backing pixmaps and
 * background color redrawing.
 *
 * Unfortunately it needs patching across all widget sets as various
 * applications expect a graphical display with common font sizes while
 * displaying appropriately sized images. Also various Gnome and custom
 * application widgets assume a common pixel sized display for its constant
 * sized padding, borders etc.
 */
#define PAVEL_MAGIC_X 1

/**
 * PAVEL_MAGIC_Y:
 *
 * Vertical number of pixels represented by one text character.
 * See #PAVEL_MAGIC_X.
 */
#define PAVEL_MAGIC_Y 1


#include "gdk/gdkprivate.h"

typedef struct _GdkFBDisplay GdkFBDisplay;
typedef struct _GdkFBWindow GdkFBWindow;

extern GdkFBWindow *gdk_root_window;
extern GdkFBDisplay *gdk_display;

#define GDK_ROOT_WINDOW()             gdk_root_window
#define GDK_DISPLAY()                 gdk_display

#define gdk_font_lookup(xid)       ((GdkFont*) (xid))

typedef enum {
  GDK_FB_0_DEGREES,
  GDK_FB_90_DEGREES,
  GDK_FB_180_DEGREES,
  GDK_FB_270_DEGREES
} GdkFBAngle;

#define GDK_FB_USE_CHILD_SHAPE ((void *)1)

/* FB specific functions: */

typedef gboolean (*GdkWindowChildChanged) (GdkWindow *window,
					   gint x, gint y,
					   gint width, gint height,
					   gpointer user_data);
typedef void     (*GdkWindowChildGetPos)  (GdkWindow *window,
					   gint *x, gint *y,
					   gpointer user_data);

void      gdk_fb_window_set_child_handler (GdkWindow              *window,
					   GdkWindowChildChanged  changed,
					   GdkWindowChildGetPos   get_pos,
					   gpointer               user_data);

void      gdk_fb_set_rotation             (GdkFBAngle angle);

#endif /* GDKFB_H */
