#include "config.h"
#include "gdkprivate-fb.h"
#include <string.h>
#include <signal.h>
#include <sys/time.h>
#include <ncurses.h>

/*
 * Reading pixel values from a generic drawable.
 */

static GetPixelRet
gdk_fb_drawable_get_color (GdkDrawable *drawable,
			   GdkGC *gc,
			   int x,
			   int y,
			   GdkColor *spot)
{
  GetPixelRet retval = GPR_NONE;
  GdkDrawableFBData *private = GDK_DRAWABLE_FBDATA (drawable);
  guchar *mem = private->mem;
  guint rowstride = private->rowstride;

  switch (private->depth)
    {
    case 4:
    case 8:
      spot->pixel = mem[x + y * rowstride];
      if (private->depth==4)
	spot->pixel&=0x0F;
      *spot = private->colormap->colors[spot->pixel];
      break;
    default:
      fprintf(stderr, "DEBUG: Dostal jsem %d\n", private->depth);
      g_assert_not_reached();
    }

  return retval;
}

/*************************************
 * gc->get_color() implementations
 *************************************/

static GetPixelRet
gdk_fb_get_color_1 (GdkDrawable *drawable,
                    GdkGC *gc,
                    int x,
                    int y,
                    GdkColor *color)
{
  GetPixelRet retval = GPR_NONE;
  GdkDrawableFBData *private = GDK_DRAWABLE_FBDATA (drawable);
  guchar *mem = private->mem;
  guint rowstride = private->rowstride;
  guchar foo;

  g_assert (private->depth == GDK_GC_FBDATA (gc)->depth);

  foo = mem[(x >> 3) + y * rowstride];
  if (foo & (1 << (x % 8)))
    *color = GDK_GC_FBDATA (gc)->values.foreground;
  else
    {
      retval = GPR_USED_BG;

      *color = GDK_GC_FBDATA (gc)->values.background;
    }

  return retval;
}


static GetPixelRet
gdk_fb_get_color_8 (GdkDrawable *drawable,
		    GdkGC *gc,
		    int x,
		    int y,
		    GdkColor *color)
{
  GdkDrawableFBData *private = GDK_DRAWABLE_FBDATA (drawable);
  guchar *mem = private->mem;
  guint rowstride = private->rowstride;
  gint pixel;

  g_assert (private->depth == GDK_GC_FBDATA (gc)->depth);
  
  pixel = mem[x + y * rowstride];
  if (private->depth==4)
    pixel&=0x0F;
  *color = private->colormap->colors[pixel];

  return GPR_NONE;
}

/*************************************
 * gc->set_pixel() implementations
 *************************************/

static void
gdk_fb_set_pixel_1(GdkDrawable *drawable,
                   GdkGC *gc,
                   int x,
                   int y,
                   gulong pixel)
{
  GdkDrawableFBData *private = GDK_DRAWABLE_FBDATA (drawable);
  guchar *mem = private->mem;
  guint rowstride = private->rowstride;
  guchar *ptr;

  g_assert (private->depth == GDK_GC_FBDATA (gc)->depth);

  ptr = mem + (y*rowstride) + (x >> 3);

  if (pixel)
    *ptr |= (1 << (x % 8));
  else
    *ptr &= ~(1 << (x % 8));
}


static void
gdk_fb_set_pixel_8(GdkDrawable *drawable,
		   GdkGC *gc,
		   int x,
		   int y,
		   gulong pixel)
{
  GdkDrawableFBData *private = GDK_DRAWABLE_FBDATA (drawable);
  guchar *mem = private->mem;
  guint rowstride = private->rowstride;

  g_assert (private->depth == GDK_GC_FBDATA (gc)->depth);

  if (private->depth==4)
    pixel&=0x0F;
  mem[x + y*rowstride] = pixel;
}


/*************************************
 * gc->fill_span() implementations
 *************************************/

static void
gdk_fb_fill_span_generic (GdkDrawable *drawable,
			  GdkGC       *gc,
			  GdkSpan     *span,
			  GdkColor    *color)
{
  int curx;
  GdkColor spot = *color;
  GdkGCFBData *gc_private;
  GdkDrawableFBData *private;
  gint left, right, y;
  int clipxoff, clipyoff; /* Amounts to add to curx & cury to get x & y in clip mask */
  int tsxoff, tsyoff;
  GdkDrawable *cmask;
  guchar *clipmem;
  guint mask_rowstride;
  GdkPixmap *ts = NULL;
  GdkDrawableFBData *ts_private;
  gboolean solid_stipple;
  GdkFunction func;

  private = GDK_DRAWABLE_FBDATA (drawable);
  gc_private = GDK_GC_FBDATA (gc);

  g_assert (gc);

  y = span->y;
  left = span->x;
  right = span->x + span->width;

  func = gc_private->values.function;
    
  cmask = gc_private->values.clip_mask;
  clipxoff = clipyoff = tsxoff = tsyoff = 0;
  mask_rowstride = 0;
  solid_stipple = FALSE;
  clipmem = NULL;
  
  if (cmask)
    {
      GdkDrawableFBData *cmask_private;
      
      cmask_private = GDK_DRAWABLE_IMPL_FBDATA (cmask);
      
      clipmem = cmask_private->mem;
      clipxoff = cmask_private->abs_x - gc_private->values.clip_x_origin - private->abs_x;
      clipyoff = cmask_private->abs_y - gc_private->values.clip_y_origin - private->abs_y;
      mask_rowstride = cmask_private->rowstride;
    }
  
  if (gc_private->values.fill == GDK_TILED &&
      gc_private->values.tile)
    {
      gint xstep;
      gint relx, rely;
      int drawh;
      GdkFBDrawingContext *dc, dc_data;

      dc = &dc_data;
      
      gdk_fb_drawing_context_init (dc, drawable, gc, FALSE, TRUE);
      
      ts = gc_private->values.tile;
      ts_private = GDK_DRAWABLE_IMPL_FBDATA (ts);
      
      rely = y - private->abs_y;
      drawh = (rely - gc_private->values.ts_y_origin) % ts_private->height;
      if (drawh < 0)
	drawh += ts_private->height;
      
      for (curx = left; curx < right; curx += xstep)
	{
	  int draww;
	  
	  relx = curx - private->abs_x;
	  
	  draww = (relx - gc_private->values.ts_x_origin) % ts_private->width;
	  if (draww < 0)
	    draww += ts_private->width;
	  
	  xstep = MIN (ts_private->width - draww, right - relx);
	  
	  gdk_fb_draw_drawable_3 (drawable, gc, GDK_DRAWABLE_IMPL (ts),
				  dc,
				  draww, drawh,
				  relx, rely,
				  xstep, 1);
	}
      
      gdk_fb_drawing_context_finalize (dc);
      
      return;
    }
  else if ((gc_private->values.fill == GDK_STIPPLED ||
	    gc_private->values.fill == GDK_OPAQUE_STIPPLED) &&
	   gc_private->values.stipple)
    {
      ts = gc_private->values.stipple;
      tsxoff = - GDK_DRAWABLE_IMPL_FBDATA (ts)->abs_x - gc_private->values.ts_x_origin - private->abs_x;
      tsyoff = - GDK_DRAWABLE_IMPL_FBDATA (ts)->abs_y - gc_private->values.ts_y_origin - private->abs_y;
      solid_stipple = (gc_private->values.fill == GDK_OPAQUE_STIPPLED);
    }

  for (curx = left; curx < right; curx++)
    {
      int maskx = curx+clipxoff, masky = y + clipyoff;
      guchar foo;
      
      if (cmask)
	{
	  foo = clipmem[masky*mask_rowstride + (maskx >> 3)];
	  
	  if (!(foo & (1 << (maskx % 8))))
	    continue;
	}
      
      if (func == GDK_INVERT)
	{
	  (gc_private->get_color) (drawable, gc, curx, y, &spot);
	  spot.pixel = ~spot.pixel;
	  spot.red = ~spot.red;
	  spot.green = ~spot.green;
	  spot.blue = ~spot.blue;
	}
      else if (func == GDK_XOR)
	{
	  (gc_private->get_color) (drawable, gc, curx, y, &spot);
	  spot.pixel ^= gc_private->values.foreground.pixel;
	}
      else if (func != GDK_COPY)
	{
	  g_warning ("Unsupported GdkFunction %d\n", func);
	}
      else if (ts)
	{
	  int wid, hih;
	  
	  ts_private = GDK_DRAWABLE_IMPL_FBDATA (ts);
	  
	  wid = ts_private->width;
	  hih = ts_private->height;
	  
	  maskx = (curx+tsxoff)%wid;
	  masky = (y+tsyoff)%hih;
	  if (maskx < 0)
	    maskx += wid;
	  if (masky < 0)
	    masky += hih;
	  
	  foo = ts_private->mem[(maskx >> 3) + ts_private->rowstride*masky];
	  if (foo & (1 << (maskx % 8)))
	    {
	      spot = gc_private->values.foreground;
	    }
	  else if (solid_stipple)
	    {
	      spot = gc_private->values.background;
	    }
	  else
	    continue;
	}
      
      (gc_private->set_pixel) (drawable, gc, curx, y, spot.pixel);

      _gdk_fb_draw_mem_cursed (drawable, curx, y, FALSE, 0, &spot, &spot);
    }
}


/*************************************
 * gc->draw_drawable() implementations
 *************************************/

static gboolean
_gdk_fb_get_drawable_bg_color (GdkDrawable *drawable,
			       GdkColor    *bg_color_return)
{
  GdkDrawableFBData *private;

  g_return_val_if_fail (drawable        != NULL, FALSE);
  g_return_val_if_fail (bg_color_return != NULL, FALSE);

  private = GDK_DRAWABLE_FBDATA (drawable);

  if (GDK_IS_WINDOW (private->wrapper))
    {
      *bg_color_return = GDK_WINDOW_P (private->wrapper)->bg_color;
      return TRUE;
    }
  if (private->colormap)
    {
      gdk_color_black (private->colormap, bg_color_return);
      return TRUE;
    }
  return FALSE;
}

static GdkGC *
_gdk_fb_get_drawables_gc (GdkDrawable *drawable_new_gc,
			  GdkDrawable *drawable_lookup,
                          GdkGC       *gc)
{
  GdkGCValues values;
  GdkGCValuesMask values_mask;

  g_return_val_if_fail (drawable_new_gc != NULL, NULL);
  g_return_val_if_fail (drawable_lookup != NULL, NULL);
  g_return_val_if_fail (gc              != NULL, NULL);

  values_mask = 0;
  if (GDK_GC_FBDATA (gc)->values_mask & GDK_GC_BACKGROUND)
    {
      values.foreground = GDK_GC_FBDATA (gc)->values.background;
      values_mask |= GDK_GC_FOREGROUND;
    }
  else if (_gdk_fb_get_drawable_bg_color (drawable_new_gc, &values.foreground))
    values_mask |= GDK_GC_FOREGROUND;
  else if (_gdk_fb_get_drawable_bg_color (drawable_lookup, &values.foreground))
    values_mask |= GDK_GC_FOREGROUND;

  if (values_mask & GDK_GC_FOREGROUND)
    {
      values.background = values.foreground;
      values_mask |= GDK_GC_BACKGROUND;
    }

  values.subwindow_mode = GDK_INCLUDE_INFERIORS;
  values_mask |= GDK_GC_SUBWINDOW;

  return gdk_gc_new_with_values (drawable_new_gc, &values, values_mask);
}

static void
gdk_fb_draw_drawable_generic (GdkDrawable *drawable,
			      GdkGC       *gc,
			      GdkPixmap   *src,
			      GdkFBDrawingContext *dc,
			      gint         start_y,
			      gint         end_y,
			      gint         start_x,
			      gint         end_x,
			      gint         src_x_off,
			      gint         src_y_off,
			      gint         draw_direction)
{
  GdkDrawableFBData *dest_private = GDK_DRAWABLE_FBDATA (drawable);
  GdkDrawableFBData *src_private = GDK_DRAWABLE_FBDATA (src);
  GdkPixmap *self_pixmap = NULL;
  
  int cur_x, cur_y;
  
  /* Make a copy of the source area if we overlap.
   * It was solved for the pixels by 'draw_direction' flag
   * but the situation of characters merging is too complex to bother with it.
   */
  if (drawable == GDK_DRAWABLE(src))
    {
     /* FIXME: Should we have our private 'GdkFBDrawingContext'? */
      GdkGC *self_pixmap_gc;

      self_pixmap = gdk_pixmap_new (GDK_DRAWABLE(src),
				    end_x - start_x, end_y - start_y,	/* width, height */
				    -1);	/* depth */

      self_pixmap_gc = _gdk_fb_get_drawables_gc(GDK_DRAWABLE_IMPL(self_pixmap), GDK_DRAWABLE(src), gc);
      gdk_draw_drawable (GDK_DRAWABLE(self_pixmap), self_pixmap_gc, src,
			 start_x + src_x_off - src_private->abs_x, start_y + src_y_off - src_private->abs_y,
			 0, 0,
			 end_x - start_x, end_y - start_y);
      g_object_unref (self_pixmap_gc);

      src = GDK_DRAWABLE_IMPL(self_pixmap);
      src_private = GDK_DRAWABLE_FBDATA (src);
      src_x_off = -start_x;
      src_y_off = -start_y;
    }

  /* Pre-clear the target area.
   * Stolen from gdk_window_clear_area()->_gdk_windowing_window_clear_area().
   * Although we try hard to guess the background color we will fail to do it.
   * We try to copy the source background color by 'Copy the background color' below.
   */
  {
    GdkGC *dest_gc;

    dest_gc = _gdk_fb_get_drawables_gc(GDK_DRAWABLE(drawable), GDK_DRAWABLE(src), gc);
    gdk_draw_rectangle (drawable, dest_gc, TRUE,
			start_x - dest_private->abs_x, start_y - dest_private->abs_y,
			end_x - start_x, end_y - start_y);
    gdk_gc_unref (dest_gc);
  }

  if (draw_direction < 0)
    {
      int tmp;
      tmp = start_y;
      start_y = end_y;
      end_y = tmp;
      start_y--;
      end_y--;
      
      tmp = start_x;
      start_x = end_x;
      end_x = tmp;
      start_x--;
      end_x--;
    }

  for (cur_y = start_y; cur_y != end_y; cur_y+=draw_direction)
    {
      for (cur_x = start_x; cur_x != end_x; cur_x+=draw_direction)
	{
	  GdkColor spot;
	  struct GdkFB_mem_cursed *ukazovatko_src;
	  
	  if (GDK_GC_FBDATA(gc)->values.clip_mask)
	    {
	      int maskx = cur_x + dc->clipxoff, masky = cur_y + dc->clipyoff;
	      guchar foo;
	      
	      foo = dc->clipmem[masky*dc->clip_rowstride + (maskx >> 3)];
	      
	      if (!(foo & (1 << (maskx % 8))))
		continue;
	    }
	  
	  switch (gdk_fb_drawable_get_color (src, gc, cur_x + src_x_off, cur_y + src_y_off, &spot))
	    {
	    case GPR_AA_GRAYVAL:
	      {
	        g_assert_not_reached();
	      }
	      break;
	    case GPR_USED_BG:
	      if (!dc->draw_bg)
		continue;
	      break;
	    case GPR_NONE:
	      break;
	    default:
	      g_assert_not_reached ();
	      break;
	    }
	  
	  (GDK_GC_FBDATA (gc)->set_pixel) (drawable, gc, cur_x, cur_y, spot.pixel);

	  ukazovatko_src = _gdk_fb_get_mem_cursed (src, cur_x + src_x_off, cur_y + src_y_off, FALSE);
	  if (ukazovatko_src)
	    _gdk_fb_draw_mem_cursed (drawable, cur_x, cur_y, FALSE,
				     ukazovatko_src->pango_glyph,
				     &ukazovatko_src->foreground, &ukazovatko_src->background);
	}
    }

  if (self_pixmap)
    g_object_unref(self_pixmap);
}


/*************************************
 * gc->fill_rectangle() implementations
 *************************************/

void
gdk_fb_fill_rectangle_generic (GdkDrawable    *drawable,
			       GdkGC          *gc,
			       GdkRectangle   *rect,
			       GdkColor       *color)
{
  GdkDrawableFBData *private;
  GdkSpan *spans;
  int i;

  private = GDK_DRAWABLE_FBDATA (drawable);

  spans = g_new (GdkSpan, rect->height);
  for (i = 0; i < rect->height; i++)
    {
      spans[i].x = rect->x - private->abs_x;
      spans[i].y = rect->y+i - private->abs_y;
      spans[i].width = rect->width;
    }
  gdk_fb_fill_spans (drawable, gc, spans, rect->height, TRUE);
  g_free (spans);
}


/*************************************
 * GC state calculation
 *************************************/

void
_gdk_fb_gc_calc_state (GdkGC           *gc,
		       GdkGCValuesMask  changed)
{
  GdkGCFBData *gc_private;
  int i;

  gc_private = GDK_GC_FBDATA (gc);

  gc_private->fill_span = gdk_fb_fill_span_generic;
  gc_private->fill_rectangle = gdk_fb_fill_rectangle_generic;

  for (i = 0; i < GDK_NUM_FB_SRCBPP; i++)
    gc_private->draw_drawable[i] = gdk_fb_draw_drawable_generic;
  
  if (changed & _GDK_FB_GC_DEPTH)  /* FIXME - proc dostava
				    * jednicku a odkud??? */
    switch (gc_private->depth)
      {
      case 1:
        gc_private->set_pixel = gdk_fb_set_pixel_1;
        gc_private->get_color = gdk_fb_get_color_1;
      case 4:
      case 8:
	gc_private->set_pixel = gdk_fb_set_pixel_8;
	gc_private->get_color = gdk_fb_get_color_8;
	break;
      default:
	g_assert_not_reached ();
	break;
      }

}


/**
 * _gdk_fb_get_mem_cursed:
 * @drawable_impl: #GdkDrawable for target drawable-impl
 * @x_abs: Absolute target coordinate;
 * 	you may need to add %GDK_DRAWABLE_FBDATA %(drawable_impl)->abs_x
 * @y_abs: Absolute target coordinate;
 * 	you may need to add %GDK_DRAWABLE_FBDATA %(drawable_impl)->abs_y
 * @offset_match: If @x_abs and @y_abs must match the target
 * 	#offset_x and #offset_y to write the attributes.
 * 	If not set the target #offset_x and #offset_y will be always overwritten.
 *
 * Internal function to get address of text character cell #GdkFB_mem_cursed.
 *
 * Returns: #GdkFB_mem_cursed pointer if
 * @x_abs and @y_abs were valid coords inside the window and @offset_match was complied.
 */
struct GdkFB_mem_cursed *
_gdk_fb_get_mem_cursed (GdkDrawable *drawable_impl,
			gint         x_abs,
			gint         y_abs,
			gboolean     offset_match)
{
  int poloha_x, poloha_y;		/* characters */
  GdkDrawableFBData *private;
  struct GdkFB_mem_cursed *ukazovatko;
  
  g_return_val_if_fail (drawable_impl != NULL, NULL);

  private = GDK_DRAWABLE_FBDATA (drawable_impl);
  
  poloha_x = x_abs / PAVEL_MAGIC_X;
  poloha_y = y_abs / PAVEL_MAGIC_Y;

  if (!((poloha_x >= (private->abs_x                  )/PAVEL_MAGIC_X) && 
        (poloha_x <= (private->abs_x+private->width -1)/PAVEL_MAGIC_X) && 
        (poloha_y >= (private->abs_y                  )/PAVEL_MAGIC_Y) && 
        (poloha_y <= (private->abs_y+private->height-1)/PAVEL_MAGIC_Y)))
    return NULL;

  ukazovatko = &private->mem_cursed[poloha_x+(private->rowstride+PAVEL_MAGIC_X-1)/PAVEL_MAGIC_X*poloha_y];

  if (offset_match && (ukazovatko->offset_x != x_abs % PAVEL_MAGIC_X
                    || ukazovatko->offset_y != y_abs % PAVEL_MAGIC_Y))
    return NULL;

  return ukazovatko;
}

/**
 * _gdk_fb_draw_mem_cursed:
 * @drawable_impl: #GdkDrawable for target drawable-impl
 * @x_abs: Absolute target coordinate;
 * 	you may need to add %GDK_DRAWABLE_FBDATA %(drawable_impl)->abs_x
 * @y_abs: Absolute target coordinate;
 * 	you may need to add %GDK_DRAWABLE_FBDATA %(drawable_impl)->abs_y
 * @offset_match: If @x_abs and @y_abs must match the target
 * 	#offset_x and #offset_y to write the attributes.
 * 	If not set the target #offset_x and #offset_y will be always overwritten.
 * @pango_glyph: Target character to write; Use %-1 to leave there the current one
 * @foreground: #GdkColor to write; Use %NULL to leave there the current one
 * @background: #GdkColor to write; Use %NULL to leave there the current one
 *
 * Internal function to write text character and/or attributes to ncurses
 * drawable. See also _gdk_fb_get_mem_cursed().
 *
 * Returns: #GdkFB_mem_cursed pointer if anything was drawn - therefore
 * @x_abs and @y_abs were valid coords inside the window and @offset_match was complied.
 * The character may not be drawn to the visible screen to return a valid pointer.
 */
struct GdkFB_mem_cursed *
_gdk_fb_draw_mem_cursed (GdkDrawable    *drawable_impl,
			 gint            x_abs,
			 gint            y_abs,
			 gboolean        offset_match,
			 PangoGlyph      pango_glyph,
			 const GdkColor *foreground,
			 const GdkColor *background)
{
  int poloha_x, poloha_y;		/* characters */
  GdkDrawableFBData *private;
  struct GdkFB_mem_cursed *ukazovatko;
  short cislo_parecku;
  chtype ch_out;
  
  g_return_val_if_fail (drawable_impl != NULL, NULL);

  private = GDK_DRAWABLE_FBDATA (drawable_impl);

  ukazovatko = _gdk_fb_get_mem_cursed(drawable_impl, x_abs, y_abs, offset_match);
  if (!ukazovatko)
    return NULL;

  if (pango_glyph != (PangoGlyph)-1)
    ukazovatko->pango_glyph = pango_glyph;
  if (foreground)
    ukazovatko->foreground = *foreground;
  if (background)
    ukazovatko->background = *background;
  if (!offset_match)
    {
      ukazovatko->offset_x = x_abs % PAVEL_MAGIC_X;
      ukazovatko->offset_y = y_abs % PAVEL_MAGIC_Y;
    }

  if (private->mem != GDK_DRAWABLE_IMPL_FBDATA (_gdk_parent_root)->mem)
    return ukazovatko;

  poloha_x = x_abs / PAVEL_MAGIC_X;
  poloha_y = y_abs / PAVEL_MAGIC_Y;

  switch (ukazovatko->pango_glyph)
    {
      case 0:
	ch_out = ' '; break;

      case 0x250c:	/* BOX DRAWINGS LIGHT DOWN AND RIGHT */
      case 0x2552:	/* BOX DRAWINGS DOWN SINGLE AND RIGHT DOUBLE */
      case 0x2553:	/* BOX DRAWINGS DOWN DOUBLE AND RIGHT SINGLE */
      case 0x2554:	/* BOX DRAWINGS DOUBLE DOWN AND RIGHT */
	ch_out = ACS_ULCORNER; break;

      case 0x2514:	/* BOX DRAWINGS LIGHT UP AND RIGHT */
      case 0x2558:	/* BOX DRAWINGS UP SINGLE AND RIGHT DOUBLE */
      case 0x2559:	/* BOX DRAWINGS UP DOUBLE AND RIGHT SINGLE */
      case 0x255a:	/* BOX DRAWINGS DOUBLE UP AND RIGHT */
	ch_out = ACS_LLCORNER; break;

      case 0x2510:	/* BOX DRAWINGS LIGHT DOWN AND LEFT */
      case 0x2555:	/* BOX DRAWINGS DOWN SINGLE AND LEFT DOUBLE */
      case 0x2556:	/* BOX DRAWINGS DOWN DOUBLE AND LEFT SINGLE */
      case 0x2557:	/* BOX DRAWINGS DOUBLE DOWN AND LEFT */
	ch_out = ACS_URCORNER; break;

      case 0x2518:	/* BOX DRAWINGS LIGHT UP AND LEFT */
      case 0x255b:	/* BOX DRAWINGS UP SINGLE AND LEFT DOUBLE */
      case 0x255c:	/* BOX DRAWINGS UP DOUBLE AND LEFT SINGLE */
      case 0x255d:	/* BOX DRAWINGS DOUBLE UP AND LEFT */
	ch_out = ACS_LRCORNER; break;

      case 0x251c:	/* BOX DRAWINGS LIGHT VERTICAL AND RIGHT */
      case 0x255e:	/* BOX DRAWINGS VERTICAL SINGLE AND RIGHT DOUBLE */
      case 0x255f:	/* BOX DRAWINGS VERTICAL DOUBLE AND RIGHT SINGLE */
      case 0x2560:	/* BOX DRAWINGS DOUBLE VERTICAL AND RIGHT */
	ch_out = ACS_LTEE;     break;

      case 0x2524:	/* BOX DRAWINGS LIGHT VERTICAL AND LEFT */
      case 0x2561:	/* BOX DRAWINGS VERTICAL SINGLE AND LEFT DOUBLE */
      case 0x2562:	/* BOX DRAWINGS VERTICAL DOUBLE AND LEFT SINGLE */
      case 0x2563:	/* BOX DRAWINGS DOUBLE VERTICAL AND LEFT */
	ch_out = ACS_RTEE;     break;

      case 0x2534:	/* BOX DRAWINGS LIGHT UP AND HORIZONTAL */
      case 0x2567:	/* BOX DRAWINGS UP SINGLE AND HORIZONTAL DOUBLE */
      case 0x2568:	/* BOX DRAWINGS UP DOUBLE AND HORIZONTAL SINGLE */
      case 0x2569:	/* BOX DRAWINGS DOUBLE UP AND HORIZONTAL */
	ch_out = ACS_BTEE;     break;

      case 0x252c:	/* BOX DRAWINGS LIGHT DOWN AND HORIZONTAL */
      case 0x2564:	/* BOX DRAWINGS DOWN SINGLE AND HORIZONTAL DOUBLE */
      case 0x2565:	/* BOX DRAWINGS DOWN DOUBLE AND HORIZONTAL SINGLE */
      case 0x2566:	/* BOX DRAWINGS DOUBLE DOWN AND HORIZONTAL */
	ch_out = ACS_TTEE;     break;

      case 0x2500:	/* BOX DRAWINGS LIGHT HORIZONTAL */
      case 0x2550:	/* BOX DRAWINGS DOUBLE HORIZONTAL */
	ch_out = ACS_HLINE;    break;

      case 0x2502:	/* BOX DRAWINGS LIGHT VERTICAL */
      case 0x2551:	/* BOX DRAWINGS DOUBLE VERTICAL */
	ch_out = ACS_VLINE;    break;

      case 0x2191:	/* UP ARROW */
	ch_out = ACS_UARROW;   break;

      case 0x2193:	/* DOWN ARROW */
	ch_out = ACS_DARROW;   break;
                                                /* for Cursed theme */
      case 0x2591:	/* LIGHT SHADE */
	ch_out = ACS_BOARD;    break;
	
      case 0x2592:	/* MEDIUM SHADE */
      case 0x2593:	/* DARK SHADE */
	ch_out = ACS_CKBOARD;    break;

      default:
	if ((ukazovatko->pango_glyph >= 0) && (ukazovatko->pango_glyph <= 0xFF))
	  ch_out = ukazovatko->pango_glyph;
        else
	  ch_out = '?';
    }

  cislo_parecku = (ukazovatko->background.pixel%8)*8+(ukazovatko->foreground.pixel%8);
  if (cislo_parecku==0)	/* ==8*COLOR_BLACK+COLOR_BLACK */
    {
      /* Do not: color_set(0, NULL);
       * as it would not draw _anything_.
       */
      cislo_parecku=8*COLOR_BLACK+COLOR_WHITE;
      ch_out = ' ';
    }

  if (ukazovatko->foreground.pixel < 8)
    color_set(cislo_parecku, NULL); 
  else 
    color_set(cislo_parecku|A_BOLD, NULL); 

  mvaddch(poloha_y, poloha_x, ch_out);

  gdk_obchodni_rejstrik_obcerstvovny();

  return ukazovatko;
}
